/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.channel;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.net.Socket;

/**
 * Class returned by
 * {@link DestinationSocketManager#getConnection(java.net.InetAddress, int)}.
 * It contains a socket, and can be used to retreive associated
 * {@link ObjectOutput }and {@link ObjectInput}.
 */
public class ConnectionState
{
  protected Socket             socket;
  protected ObjectInputStream  input;
  protected ObjectOutputStream output;

  /**
   * Creates a new connection state for the specified socket.
   * 
   * @param socket a socket.
   */
  public ConnectionState(Socket socket)
  {
    this.socket = socket;
  }

  /**
   * Returns <code>true</code> if the socket is closed or is deconnected.
   * 
   * @return <code>true</code> if the socket is closed or is deconnected.
   */
  public boolean isClosed()
  {
    return socket.isClosed() || !socket.isConnected();
  }

  /**
   * Returns an {@link ObjectInput }that can be used to read data from the
   * socket.
   * 
   * @return an {@link ObjectInput }connected to the socket.
   * @throws IOException if the {@link ObjectInput }can't be created.
   */
  public ObjectInput getInput() throws IOException
  {
    if (input == null)
    {
      input = new ObjectInputStream(socket.getInputStream());
    }
    return input;
  }

  /**
   * Returns an {@link ObjectOutput }that can be used to write data to the socket.
   * @return an {@link ObjectOutput }connected to the socket. 
   * @throws IOException if the {@link ObjectOutput }can't be created.
   */
  public ObjectOutput getOutput() throws IOException
  {
    if (output == null)
    {
      output = new ObjectOutputStream(socket.getOutputStream());
    }
    return output;
  }

  /**
   * Closes the connection.
   */
  public void close()
  {
    if (input != null)
    {
      try
      {
        input.close();
      }
      catch (IOException e)
      {
        // ignore
      }
    }
    if (output != null)
    {
      try
      {
        output.close();
      }
      catch (IOException e)
      {
        // ignore
      }
    }
    if (!socket.isClosed())
    {
      try
      {
        socket.close();
      }
      catch (IOException e)
      {
        // ignore
      }
    }
  }
}