/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.multiplexer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InitializationException;
import org.objectweb.dream.Pull;
import org.objectweb.dream.Push;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.time.SetTimeStamp;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * Implementation of the <code>PullPushMultiplexer</code> interface. This
 * multiplexers periodically pulls its inputs according to the parameters given
 * to the attach method.
 */
public class PullPushPeriodicMultiplexerImpl extends AbstractComponent
    implements
      PullPushMultiplexer
{
  /** The registered attachments. */
  protected ArrayList      attachments;
  /** An hashtable storing the intputs of this component. */
  protected Map            inputs;                                 // name ->
                                                                   // input
  /** An hashtable storing the outputs of this component. */
  protected Map            outputs;                                // name ->
                                                                   // output
  /** The message manager client interface of this component */
  protected MessageManager messageManagerItf;
  /** The SetTimeStamp client interface of this component. */
  protected SetTimeStamp   setTimeStamp;
  /** An integer used to generate attachment ids. */
  protected int            attachmentId = 0;

  private Task             pumpTask     = new PumpTask("pumpTask");

  /**
   * Constructor.
   */
  public PullPushPeriodicMultiplexerImpl()
  {
    this.attachments = new ArrayList();
    this.inputs = new Hashtable();
    this.outputs = new Hashtable();
  }

  // -------------------------------------------------------------------------
  // Implementation of the PullPushMultiplexer interface
  // -------------------------------------------------------------------------

  /**
   * Attaches a set of inputs to a set of outputs.
   * 
   * @param inputNames the names of the inputs to be attached. These inputs must
   *          have been previously bound (with the same name) using the
   *          <code>BindingController</code> interface.
   * @param inputContexts the contexts to be passed when pulling the inputs.
   * @param outputNames the names of the outputs to wich intputs must be
   *          attached. These outputs must have been previously bound (with the
   *          same name) using the <code>BindingController</code> interface.
   * @param outputContexts the contexts to be passed when pushing the outputs.
   * @param parameters the attachement parameters. The <code>parameters</code>
   *          Map can have the following <code>String</code> as keys:
   *          <ul>
   *          <li><i>startingDate </i>: mapped to a <code>Long</code> that
   *          specifies in how many ms should the attachement be started. (0
   *          means now).</li>
   *          <li><i>endDate </i>: mapped to a <code>Long</code> that
   *          specifies in how many ms should the attachement be started. (0
   *          means now).</li>
   *          <li><i>pullingFrequency </i>: mapped to a <code>Long</code>
   *          that specifies the time to sleep between two pulling of the
   *          attached inputs.</li>
   *          <li><i>endDate </i>: mapped to a <code>Long</code> that
   *          specifies in how many ms should the attachement be stop. (must be >
   *          startingDate)</li>
   *          </ul>
   * @return an <code>Attachment</code> object
   * @throws NoSuchInterfaceException if inputNames or outputNames contains the
   *           name of an interface to which the component has not been bound.
   * @see PeriodicAttachment
   * @see org.objectweb.dream.multiplexer.PullPushMultiplexer#attach(String[],
   *      Map[], String[], Map[], Map)
   */
  public Attachment attach(String[] inputNames, Map[] inputContexts,
      String[] outputNames, Map[] outputContexts, Map parameters)
      throws NoSuchInterfaceException
  {

    long startingDate = 0;
    if (parameters.get(PeriodicAttachment.STARTING_DATE) != null)
    {
      startingDate = ((Long) parameters.get(PeriodicAttachment.STARTING_DATE))
          .longValue();
    }
    long endDate = 0;
    if (parameters.get(PeriodicAttachment.END_DATE) != null)
    {
      endDate = ((Long) parameters.get(PeriodicAttachment.END_DATE))
          .longValue();
    }
    long pullingFrequency = 0;
    if (parameters.get(PeriodicAttachment.PULLING_FREQUENCY) != null)
    {
      pullingFrequency = ((Long) parameters
          .get(PeriodicAttachment.PULLING_FREQUENCY)).longValue();
    }
    Pull[] attachmentInputs = new Pull[inputs.size()];
    for (int i = 0; i < inputNames.length; i++)
    {
      attachmentInputs[i] = (Pull) inputs.get(inputNames[i]);
    }
    Push[] attachmentOutputs = new Push[outputs.size()];
    for (int i = 0; i < outputNames.length; i++)
    {
      attachmentOutputs[i] = (Push) outputs.get(outputNames[i]);
    }
    PeriodicAttachment attachment = new PeriodicAttachment(attachmentId++,
        inputNames, attachmentInputs, inputContexts, outputNames,
        attachmentOutputs, outputContexts, startingDate, pullingFrequency,
        endDate);
    synchronized (attachments)
    {
      attachments.add(attachment);
      attachments.notifyAll();
    }
    // Clone does not clone the inputs and outputs arrays
    return (Attachment) attachment.clone();

  }

  /**
   * @see PullPushMultiplexer#detach(Attachment)
   */
  public void detach(Attachment attachment)
  {
    synchronized (attachments)
    {
      // TODO the following instruction must not be necessary
      ((PeriodicAttachment) attachment).setEndDate(System.currentTimeMillis());
      attachments.remove(attachment);
    }
  }

  /**
   * @see PullPushMultiplexer#update(Attachment)
   */
  public void update(Attachment attachment)
  {
    synchronized (attachments)
    {
      attachments.remove(attachment);
      Attachment newAttachment = (Attachment) attachment.clone();
      // Set the inputs and outputs
      Pull[] attachmentInputs = new Pull[inputs.size()];
      for (int i = 0; i < newAttachment.inputNames.length; i++)
      {
        attachmentInputs[i] = (Pull) inputs.get(newAttachment.inputNames[i]);
      }
      Push[] attachmentOutputs = new Push[outputs.size()];
      for (int i = 0; i < newAttachment.outputNames.length; i++)
      {
        attachmentOutputs[i] = (Push) outputs.get(newAttachment.outputNames[i]);
      }
      newAttachment.setInputs(attachmentInputs);
      newAttachment.setOutputs(attachmentOutputs);
      attachments.add(newAttachment);
    }
  }

  /**
   * @see PullPushMultiplexer#getAttachments()
   */
  public ArrayList getAttachments()
  {
    ArrayList returnedAttachments = new ArrayList();
    Iterator iter = attachments.iterator();
    while (iter.hasNext())
    {
      returnedAttachments.add(((PeriodicAttachment) iter.next()).clone());
    }
    return returnedAttachments;
  }

  // -------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // -------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.startsWith(Push.OUT_PUSH_ITF_NAME))
    {
      outputs.put(clientItfName, serverItf);
    }
    else if (clientItfName.startsWith(Pull.IN_PULL_ITF_NAME))
    {
      inputs.put(clientItfName, serverItf);
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(SetTimeStamp.ITF_NAME))
    {
      setTimeStamp = (SetTimeStamp) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(String)
   */
  public synchronized void unbindFc(String clientItfName)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.unbindFc(clientItfName);
    if (clientItfName.startsWith(Push.OUT_PUSH_ITF_NAME))
    {
      outputs.remove(clientItfName);
    }
    else if (clientItfName.startsWith(Pull.IN_PULL_ITF_NAME))
    {
      inputs.remove(clientItfName);
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public synchronized String[] listFc()
  {
    int size = outputs.size() + inputs.size() + 2;
    String[] tab = new String[size];
    outputs.keySet().toArray(tab);
    String[] inputsArr = new String[inputs.size()];
    inputs.keySet().toArray(inputsArr);
    int j = 0;
    for (int i = outputs.size(); i < outputs.size() + inputs.size(); i++)
    {
      tab[i] = inputsArr[j];
      j++;
    }
    tab[inputs.size() + outputs.size()] = MessageManager.ITF_NAME;
    tab[inputs.size() + outputs.size() + 1] = SetTimeStamp.ITF_NAME;
    return tab;
  }

  // ---------------------------------------------------------------------------
  // Overriding of initComponent method
  // ---------------------------------------------------------------------------

  protected void initComponent(Component componentItf)
      throws InitializationException
  {
    super.initComponent(componentItf);
    try
    {
      Util.addTask(componentItf, pumpTask, Collections.EMPTY_MAP);
    }
    catch (Exception e)
    {
      throw new InitializationException(e, componentItf, "Can't add task");
    }
  }

  // -------------------------------------------------------------------------
  // Private class defining the task of this component
  // -------------------------------------------------------------------------

  private class PumpTask extends AbstractTask
  {
    /**
     * @param name the name of the task.
     */
    public PumpTask(String name)
    {
      super(name);
    }

    /**
     * @see org.objectweb.dream.control.activity.task.Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      long nextWakeup = Long.MAX_VALUE;
      long currentTime = setTimeStamp.setTimeStamp();
      synchronized (attachments)
      {
        int size = attachments.size();
        for (int i = 0; i < size; i++)
        {
          PeriodicAttachment attachment = (PeriodicAttachment) attachments
              .get(i);
          long endDate = attachment.getEndDate();
          if ((endDate != 0) && (endDate < currentTime))
          { // Attachment has
            // expired,
            // remove it.
            attachments.remove(i);
            i--;
            size--;
            continue;
          }
          if (attachment.getNextDeadline() <= currentTime)
          { // We need to probe
            // this
            // one
            try
            { // Probe the resources and insert the values in the buffer
              Pull[] attachmentInputs = attachment.getInputs();
              Map[] inputContexts = attachment.getInputContexts();
              Push[] attachmentOutputs = attachment.getOutputs();
              Map[] outputContexts = attachment.getOutputContexts();
              for (int j = 0; j < attachmentInputs.length; j++)
              {
                Message m = attachmentInputs[j].pull(inputContexts[j]);
                for (int k = 0; k < attachmentOutputs.length; k++)
                {
                  // TODO add the clone and aggregate tests
                  attachmentOutputs[k].push(m, outputContexts[k]);
                }
              }
            }
            catch (Exception e)
            { // An error occured while monitoring,
              // remove
              // this attachment
              e.printStackTrace();
              attachments.remove(i);
              i--;
              size--;
              continue;
            }
            // Update the next deadline for this attachment
            attachment.setNextDeadline(currentTime
                + attachment.pullingFrequencyInMillis());
          }

          // Update next wake up time
          if (nextWakeup > attachment.getNextDeadline())
          {
            nextWakeup = attachment.getNextDeadline();
          }
        }

        // Wait until next deadline
        if (nextWakeup == Long.MAX_VALUE)
        {
          // No attachment, wait until a new attachment is added
          attachments.wait();
        }
        else
        {
          long timeToSleep = nextWakeup - currentTime;
          if (timeToSleep > 0)
          {
            attachments.wait(timeToSleep);
          }
        }
      }
      return EXECUTE_AGAIN;
    }
  }

}