/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s):
 */

package org.objectweb.dream.protocol.atomicity;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Transient impelmentation of the Reactor component. This implementation
 * guarantees that notifications emited by the reacting agent are effectuively
 * sent if the reaction doesn't throw an exception.
 */
public class AtomicReactorImpl extends AbstractComponent implements Push
{

  /**
   * The name of the client interface used to pass message received on the
   * incommingInPush interface.
   */
  public static final String INCOMMING_OUT_PUSH_ITF_NAME = "incommingOutPush";

  /**
   * The name of the client interface used to send message pulled from the
   * waiting list.
   */
  public static final String OUTGOING_OUT_PUSH_ITF_NAME  = "outgoingOutPush";

  /**
   * The name of the client interface used to retreivd notifications emited
   * during reaction. <B>This interface MUST be in non blocking mode </B>
   * 
   * @see org.objectweb.dream.queue.PushPullQueueAttributeController
   */
  public static final String WAITING_PULL_ITF_NAME       = "waitingPull";

  private Push               incommingOutPushItf;
  private Push               outgoingOutPushItf;
  private Pull               waitingPullItf;
  private SetReactorThread   setReactorThreadItf;
  private MessageManager     messageManagerItf;

  /** The previous thread that was executing a reaction. */
  private Thread             previousReactorThread       = null;

  /**
   * @see Push#push(Message, Map)
   */
  public synchronized void push(Message message, Map context)
      throws PushException
  {

    Thread currentThread = Thread.currentThread();
    if (currentThread != previousReactorThread)
    {
      previousReactorThread = currentThread;
      setReactorThreadItf.setReactorThread(currentThread);
    }

    try
    {
      try
      {
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger
              .log(BasicLevel.DEBUG, "Forward incomming message : " + message);
        }
        incommingOutPushItf.push(message, context);
      }
      catch (PushException e)
      {
        logger.log(BasicLevel.INFO,
            "Exception catched during reaction, emited messages are dropped. (message="
                + message + ")", e.getCause());

        Message msg = waitingPullItf.pull(null);
        while (msg != null)
        {
          messageManagerItf.deleteMessage(msg);
          msg = waitingPullItf.pull(null);
        }
      }

      Message msg = waitingPullItf.pull(null);
      while (msg != null)
      {
        logger.log(BasicLevel.DEBUG, "Push outgoing message : " + message);
        outgoingOutPushItf.push(msg, null);
        msg = waitingPullItf.pull(null);
      }
    }
    catch (PullException e)
    {
      throw new PushException("An error occurs while pulling waiting messages",
          e);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{OUTGOING_OUT_PUSH_ITF_NAME,
        INCOMMING_OUT_PUSH_ITF_NAME, WAITING_PULL_ITF_NAME,
        SetReactorThread.SET_REACTOR_THREAD_ITF_NAME, MessageManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(OUTGOING_OUT_PUSH_ITF_NAME))
    {
      outgoingOutPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(INCOMMING_OUT_PUSH_ITF_NAME))
    {
      incommingOutPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(WAITING_PULL_ITF_NAME))
    {
      waitingPullItf = (Pull) serverItf;
    }
    else if (clientItfName.equals(SetReactorThread.SET_REACTOR_THREAD_ITF_NAME))
    {
      setReactorThreadItf = (SetReactorThread) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

}