/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.lifecycle;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.control.lifecycle.LifeCycleCoordinator;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Life cycle mixin to be used with interceptor generated by
 * {@link LifeCycleInterceptorGenerator}.
 * <p>
 * <p>
 * <b>Requirements </b>
 * <ul>
 * <li>This mixin must be used with the
 * {@link org.objectweb.fractal.julia.UseComponentMixin }mixin.</li>
 * <li>This mixin must be used with the {@link SimpleLifeCycleControllerMixin}
 * lifecycle controller implementation.</li>
 * <li>This mixin must be used with the {@link LifeCycleInterceptorGenerator}
 * interceptor generator.</li>
 * </ul>
 */
public abstract class ThreadCounterMixin implements LifeCycleCoordinator
{

  /** The array of currently executing threads. */
  protected Thread[]             fcExecutingThreads;
  /** The array of the invocation counter for each threads. */
  protected int[]                fcInvocationCounters;
  /** The number of currently executing thread */
  protected int                  fcNbExcutingThread;

  protected LifeCycleCoordinator fcCoordinator;

  // ---------------------------------------------------------------------------
  // Fields and methods added and overriden by the mixin class
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public void initFcController(final InitializationContext ic)
      throws InstantiationException
  {

    // intialize fields
    fcExecutingThreads = new Thread[2];
    fcInvocationCounters = new int[2];
    fcNbExcutingThread = 0;

    _super_initFcController(ic);
  }

  /**
   * @see LifeCycleCoordinator#setFcStarted()
   */
  public boolean setFcStarted() throws IllegalLifeCycleException
  {
    synchronized (this)
    {
      if (_super_setFcStarted())
      {
        notifyAll();
        return true;
      }
      return false;
    }
  }

  /**
   * @see LifeCycleCoordinator#setFcStopping(LifeCycleCoordinator)
   */
  public void setFcStopping(LifeCycleCoordinator coordinator)
      throws IllegalLifeCycleException
  {
    synchronized (this)
    {
      _this_weaveableLCCLogger.log(BasicLevel.DEBUG, "Stopping component");
      _this_weaveableFcState = 1;
      fcCoordinator = coordinator;
      if (!_this_isInactivated(coordinator))
      {
        for (int i = 0; i < fcExecutingThreads.length; i++)
        {
          // interrupt executing threads
          if (fcExecutingThreads[i] != null)
          {
            fcExecutingThreads[i].interrupt();
          }
        }
      }
    }
  }

  /**
   * @see LifeCycleCoordinator#setFcStopped()
   */
  public boolean setFcStopped() throws IllegalLifeCycleException
  {
    synchronized (this)
    {
      if (_super_setFcStopped())
      {
        fcCoordinator = null;
        return true;
      }
      return false;
    }
  }

  // ---------------------------------------------------------------------------
  // Methods called by the associated interceptor
  // ---------------------------------------------------------------------------

  /**
   * Registers the current thread has an executing thread. If it is already
   * registered, increments it counter. If the component is started this method
   * just registers the current thread. If the component is stopped, it waits
   * until the component is restarted, ant then registers the current thread.
   * Finally if the component is stopping, threr are two cases. If no thread is
   * executing the component, this method asks the coordinator if the method
   * call can be executed, or if it should be delayed until the component is
   * restarted. The method then blocks or not, depending on the coordinator's
   * response, before registering the current thread.
   * 
   * @param s not used.
   * @return the index of the current thread in the array of executing threads.
   *         This value is passed to the
   *         {@link #removeCurrentThread(String, int) }to executing thread.
   */
  public int addCurrentThread(String s)
  {
    synchronized (this)
    {
      boolean ok;
      if (_this_weaveableFcState != 2)
      {
        do
        {
          if (_this_weaveableFcState == 0)
          {
            ok = false;
          }
          else if (_this_weaveableFcState == 1)
          {
            if (fcNbExcutingThread == 0)
            {
              ok = fcCoordinator.fcActivated(_this_getFcCoordinator());
            }
            else
            {
              ok = true;
            }
          }
          else
          {
            ok = true;
          }

          if (!ok)
          {
            try
            {
              wait();
            }
            catch (InterruptedException e)
            {

            }
          }
        }
        while (!ok);
      }

      return registerCurrentThread();
    }
  }

  /**
   * Unregister the current thread. Finally if the component is stoping, there
   * are two cases. If the component is stopping, and no more thread is
   * executing it, this method notifies the coordinator that the component has
   * become inactive.
   * 
   * @param s not used.
   * @param i the index of the current thread in the array of executing threads.
   */
  public void removeCurrentThread(String s, int i)
  {
    synchronized (this)
    {
      unregisterCurrentThread(i);
    }
  }

  /**
   * Registers the current thread.
   * 
   * @return the index of the current thread in the array of executing threads.
   */
  public int registerCurrentThread()
  {
    int freeIndex = -1;
    Thread currentThread = Thread.currentThread();
    for (int i = 0; i < fcExecutingThreads.length; i++)
    {
      Thread thread = fcExecutingThreads[i];
      if (thread == currentThread)
      {
        fcInvocationCounters[i]++;
        return i;
      }
      else if (thread == null)
      {
        freeIndex = i;
      }
    }
    if (freeIndex == -1)
    {
      // no space left in arrays
      Thread[] precThreads = fcExecutingThreads;
      int[] precCounters = fcInvocationCounters;
      fcExecutingThreads = new Thread[precThreads.length + 2];
      fcInvocationCounters = new int[precThreads.length + 2];

      System.arraycopy(precThreads, 0, fcExecutingThreads, 0,
          precThreads.length);
      System.arraycopy(precCounters, 0, fcInvocationCounters, 0,
          precCounters.length);
      freeIndex = precThreads.length;
    }

    fcExecutingThreads[freeIndex] = currentThread;
    fcInvocationCounters[freeIndex] = 1;
    fcNbExcutingThread++;
    return freeIndex;
  }

  /**
   * Unregisters current thread.
   * 
   * @param i the index of the current thread in the array of executing threads.
   */
  public void unregisterCurrentThread(int i)
  {
    fcInvocationCounters[i]--;
    if (fcInvocationCounters[i] == 0)
    {
      fcExecutingThreads[i] = null;
      fcNbExcutingThread--;
      _this_isInactivated(fcCoordinator);
    }
  }

  /**
   * Returns <code>true</code> If the component is stopping, and no more
   * thread is executing it and super implementation returns <code>true</code>.
   * 
   * @return <code>true</code> If the component is inactive
   */
  public boolean checkInactivity()
  {
    return _this_weaveableFcState != 2 && fcNbExcutingThread == 0
        && _super_checkInactivity();
  }

  // ---------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // ---------------------------------------------------------------------------

  /**
   * The <code>weaveableC</code> field required by this mixin. This field is
   * supposed to reference the {@link Component}interface of the component to
   * which this controller object belongs.
   */
  public Component _this_weaveableC;

  /**
   * the <code>weaveableFcState</code> field reqired by this mixin. This field
   * is supposed to reference the current state of the component lifecycle.
   * 
   * @see SimpleLifeCycleControllerMixin#weaveableFcState
   */
  public int       _this_weaveableFcState;

  /**
   * The <tt>weaveableLCCLogger</tt> field required by this mixin. This field
   * is supposed to reference the {@link Logger }of this controller.
   */
  public Logger    _this_weaveableLCCLogger;

  /**
   * The {@link LifeCycleCoordinator#setFcStarted() }method overriden by this
   * mixin.
   * 
   * @see LifeCycleCoordinator#setFcStarted()
   */
  public abstract boolean _super_setFcStarted()
      throws IllegalLifeCycleException;

  /**
   * The {@link LifeCycleCoordinator#setFcStopped() }method overriden by this
   * mixin.
   * 
   * @see LifeCycleCoordinator#setFcStopped()
   */
  public abstract boolean _super_setFcStopped()
      throws IllegalLifeCycleException;

  /**
   * The <code>checkInactivity</code> method overriden by this mixin. This
   * method is supposed to work as this
   * {@link SimpleLifeCycleControllerMixin#checkInactivity() }method.
   * 
   * @see SimpleLifeCycleControllerMixin#checkInactivity()
   */
  public abstract boolean _super_checkInactivity();

  /**
   * The <code>getFcCoordinator</code> method overriden by this mixin. This
   * method is supposed to work as this
   * {@link SimpleLifeCycleControllerMixin#getFcCoordinator() }method.
   * 
   * @see SimpleLifeCycleControllerMixin#getFcCoordinator()
   */
  public abstract LifeCycleCoordinator _this_getFcCoordinator();

  /**
   * The <code>isInactivated</code> method used by this mixin. This method is
   * supposed to work as this
   * {@link SimpleLifeCycleControllerMixin#isInactivated }method.
   * 
   * @see SimpleLifeCycleControllerMixin#isInactivated
   */
  public abstract boolean _this_isInactivated(LifeCycleCoordinator coordinator);

  /**
   * The {@link org.objectweb.fractal.julia.Controller#initFcController }method
   * overriden by this mixin.
   * 
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public abstract void _super_initFcController(InitializationContext ic)
      throws InstantiationException;
}